;MC145170ctl.ASM     G4JNt  Jan 2010

;Controls LMX23x6 familiy synth.   Boot synth from EE at turn on
;Full control via RS232 interface using port change interrupt

;  Serial Command  protocol     Rxxxx[cr]    Input new R data, no update
;                               Nxxxxx[cr]   Input new N data, no update
;                               Fxxxxx[cr]   Input new F data, no update
;                               U [cr]       Update chip with new data, don't store
;                               W [cr]       Update chip and write to EE

    PROCESSOR 12F629
    INCLUDE "P12F629.INC"
    errorlevel 1

;   ===============================================================

    #define     SerIn       GPIO, 0     ;Also PGD,  Port change enabled
    #define     SerOut      GPIO, 1     ;Also PGC
    #define     Data170     GPIO, 2     ;Pin 5
    #define     En170       GPIO, 4     ;Pin 3
    #define     Clk170      GPIO, 5     ;Pin 2

    #define     LED         GPIO, 1     ;LED on SerOut for debugging

 RS232POL   =   1         ; 1 = RS232 levels,  0 = Logic/FTDI Interface
 BAUDCONST  =   d'138'      ;Half baud delay, 1200 baud

  __config  0x3FC4      ;INTOSC, I/O on GP4/5
  ;CPD Off, CP Off, Boden On, MCLRE = I/O, PWRTE on, WDT Off,

  __idlocs  0x8065

    R = d'15'      ;
    N = d'100'     
    Cvalue   = b'01100011'      ;Phase Det A Fref/8, LD, Fv, Fr on


  org   0x2100
Rdata    de  HIGH(R), LOW(R) 	;R data (15 bits in total)
Ndata    de  HIGH(N), LOW(N)		;N data (16 bits in total)
Cdata    de  Cvalue			;C data (8 bits)

  org   0
    nop
    clrw
    movwf   INTCON          ;disable interrupts
    goto    StartUp         ;jump to main code
;================================
  org 4     ;Interrupt routine, every time RotI goes low to high
    movwf   Wreg            ;Save W
    swapf   STATUS , W      ; nb.  movwf does not affect any STATUS bits
    movwf   StsReg          ;Stave STATUS, without affecting it

    btfss   INTCON , GPIF   ;Test for port B change
    goto    CarryOnInt      ;
                            ;
    bcf     INTCON , GPIE   ;
    call    Assemble232     ;MUSTN'T Waste ANY TIME UP TO HERE !!!!

    sublw   0x0D            ;Check for [cr] terminating any command,
    btfsc   STATUS, Z       ;  otherwise chars are pushed onto stack
    call    CheckSerCmd     ;Decode the command and do any actions

    bcf     INTCON , GPIF   ;Clear the flag
    bsf     INTCON , GPIE   ;Re-enable the interrupt
;...............
CarryOnInt
    swapf   StsReg , W      ;Restore STATUS
    movwf   STATUS
    swapf   Wreg            ;Restore W
    swapf   Wreg , W        ; nb.  movf affects Z bit
    retfie
;=================================
StartUp 
    movlw   7
    movwf   CMCON
    bsf     STATUS,RP0      ;ram page 1

    movlw   b'00000001'     ;GP0 is the serial input, all others out
    movwf   TRISIO          ;
    movwf   IOC             ;Port change interrupt on GP0

    movlw   b'00000000'     ;Pull-ups enabled
	movwf	OPTION_REG

    call    0x3FF           ;Get factory stored OSCCAL value
    movwf   OSCCAL
    bcf     STATUS, RP0

    call    Init170


  if RS232POL == 1
    bcf   SerOut      ;RS232 idle state (note RS232 polarity !)
  else
    bsf   SerOut
  endif

    call    LongDelay       ;Give it some time to settle
    call    IntroMessage

    movlw   Rdata
    call    GetEE_12F
    movwf   R1
    
    movlw   Rdata + 1
    call    GetEE_12F
    movwf   R0
    
    movlw   Ndata
    call    GetEE_12F
    movwf   N1

    movlw   Ndata + 1
    call    GetEE_12F
    movwf   N0

    movlw   Cdata
    call    GetEE_12F
    movwf   Creg

    call    SendData                        
    call    SendBack

    movlw   b'10001000'      ;Port Change interrupt
    movwf   INTCON

MainLoop        ;
    goto    MainLoop

;------------------------------------
SendData                ;Now Send complete R  N and C register set   
	movf	Creg, W
	movwf	A0    
	movlw	d'8'		;8 clocks define C data
	movwf	Counter
	bcf     En170
Cloop
	rlf     A0
	btfss	STATUS, C
	bcf     Data170
	btfsc	STATUS, C
	bsf     Data170
	bsf     Clk170
	nop
	bcf     Clk170

	decfsz	Counter
	goto	Cloop
	bsf     En170

	call    Delay

	movf	R1, W
	movwf	A1
	movf	R0, W
	movwf	A0
	rlf     A0          ;left-align the relevant 15 bits
	rlf     A1

	movlw	d'15'		;15 clocks define R data (can use 24)
	movwf	Counter
	bcf     En170
Rloop
	rlf     A0
	rlf     A1
	btfss	STATUS, C
	bcf     Data170
	btfsc	STATUS, C
	bsf     Data170
	bsf     Clk170
	nop
	bcf     Clk170
	decfsz	Counter
	goto	Rloop
	bsf     En170

	call    Delay

	movf	N1, W
	movwf	A1
	movf	N0, W
	movwf	A0

	movlw	d'16'		;16 clocks define N data
	movwf	Counter
	bcf     En170
Nloop
	rlf     A0
	rlf     A1
	btfss	STATUS, C
	bcf     Data170
	btfsc	STATUS, C
	bsf     Data170
	bsf     Clk170
	nop
	bcf     Clk170
	decfsz	Counter
	goto    Nloop
	bsf     En170

	return
;------------------------------------
SendBack                ;Send contents of R and N registers on RS232 interface
    call    CrLf
    movlw   "R"
    call    SendByte
    movlw   "-"
    call    SendByte

    swapf   R1, W
    call    SendHex
    movf    R1, W
    call    SendHex

    swapf   R0, W
    call    SendHex
    movf    R0, W
    call    SendHex

    call    CrLf
    movlw   "N"
    call    SendByte
    movlw   "-"
    call    SendByte

    swapf   N1, W
    call    SendHex
    movf    N1, W
    call    SendHex

    swapf   N0, W
    call    SendHex
    movf    N0, W
    call    SendHex

    call    CrLf
    movlw   "C"
    call    SendByte
    movlw   "-"
    call    SendByte    

    swapf   Creg, W
    call    SendHex
    movf    Creg, W
    call    SendHex

    call    CrLf
    return
;------------------------------------
DecodeHex           ;Single ASCII character to binary 
    movwf   Temp            ;Uses Temp
    movlw   0x30
    subwf   Temp            ;subtract $30 and leave in Temp
    btfss   STATUS , C      ;if < $30, result -ve so C=0, error, set=0
    clrf    Temp

    movf    Temp , W        
    sublw   d'9'            ;W = 9 - W
    btfsc   STATUS , C      ;if -ve, C=0, char > 9, so subtract 7
    goto    ConvDone
    movlw   d'7'
    subwf   Temp            ;
ConvDone
    movf    Temp , W        ;
    andlw   0x0F            ;mask out any rubbish left in
    return              
;------------------------------------
SendHex         ;Send a single Nibble as RS232
    andlw   0x0F
    addlw   0x30
    movwf   Temp
    sublw   0x39
    btfsc   STATUS , C
    goto    CharDone
    movlw   7
    addwf   Temp
CharDone
    movf    Temp , W
    call    SendByte
    return
;-----------------------------------
IntroMessage
    movlw   "M"
    call    SendByte
    movlw   "C"
    call    SendByte
    movlw   "1"
    call    SendByte
    movlw   "4"
    call    SendByte
    movlw   "5"
    call    SendByte
    movlw   "1"
    call    SendByte
    movlw   "7"
    call    SendByte
    movlw   "0"
    call    SendByte
    movlw   " "
    call    SendByte    
    movlw   "C"
    call    SendByte
    movlw   "o"
    call    SendByte
    movlw   "n"
    call    SendByte
    movlw   "t"
    call    SendByte
    movlw   "r"
    call    SendByte
    movlw   "o"
    call    SendByte
    movlw   "l"
    call    SendByte

    movlw   " "
    call    SendByte
    movlw   " "
    call    SendByte
    movlw   " "
    call    SendByte
    movlw   " "
    call    SendByte


    movlw   "G"
    call    SendByte
    movlw   "4"
    call    SendByte
    movlw   "J"
    call    SendByte
    movlw   "N"
    call    SendByte
    movlw   "T"
    call    SendByte
    call    CrLf

    movlw   "C"
    call    SendByte
    movlw   "o"
    call    SendByte
    movlw   "m"
    call    SendByte
    movlw   "m"
    call    SendByte
    movlw   "a"
    call    SendByte
    movlw   "n"
    call    SendByte
    movlw   "d"
    call    SendByte
    movlw   "s"
    call    SendByte
    movlw   ":"
    call    SendByte
    call    CrLf

    movlw   "R"
    call    SendByte
    movlw   "x"
    call    SendByte
    movlw   "x"
    call    SendByte
    movlw   "x"
    call    SendByte
    movlw   "x"
    call    SendByte
    call    CrLf

    movlw   "N"
    call    SendByte
    movlw   "x"
    call    SendByte
    movlw   "x"
    call    SendByte
    movlw   "x"
    call    SendByte
    movlw   "x"
    call    SendByte
   
    call    CrLf

    movlw   "C"
    call    SendByte
    movlw   "x"
    call    SendByte
    movlw   "x"
    call    SendByte
   
   
    call    CrLf

    movlw   "["
    call    SendByte
    movlw   "U"
    call    SendByte
    movlw   "]"
    call    SendByte
    movlw   "p"
    call    SendByte
    movlw   "d"
    call    SendByte
    movlw   "a"
    call    SendByte
    movlw   "t"
    call    SendByte
    movlw   "e"
    call    SendByte
    call    CrLf

    movlw   "["
    call    SendByte
    movlw   "W"
    call    SendByte
    movlw   "]"
    call    SendByte
    movlw   "r"
    call    SendByte
    movlw   "i"
    call    SendByte
    movlw   "t"
    call    SendByte
    movlw   "e"
    call    SendByte
    movlw   " "
    call    SendByte
    movlw   "E"
    call    SendByte
    movlw   "E"
    call    SendByte

    call    CrLf

    return
;-----------------------------------
CrLf
    movlw   0x0D
    call    SendByte
    movlw   0x0A
    call    SendByte
    return
;---------------------------------
SendByte                    
	movwf   Temp
 if RS232POL == 1
    bsf     SerOut
 else
    bcf     SerOut             ;start bit
 endif
	call    HalfBitDelay        ;
	call    HalfBitDelay
	movlw   8
	movwf   Counter
SendDatLoop
	rrf     Temp                ;bit into C

 if RS232POL == 1
	btfsc   STATUS , C
    bcf     SerOut
	btfss   STATUS , C
    bsf     SerOut             ;could have 2us jitter depending on 1/0
 else
	btfsc   STATUS , C
    bsf     SerOut
	btfss   STATUS , C
    bcf     SerOut             ;could have 2us jitter depending on 1/0
 endif

	call    HalfBitDelay
	call    HalfBitDelay
	decfsz  Counter
	goto    SendDatLoop         ;loop 8 + BitDelay long

 if RS232POL == 1
    bcf     SerOut
 else
    bsf     SerOut
 endif
	call    HalfBitDelay    
	call    HalfBitDelay        ;Two stop bits
	call    HalfBitDelay
	call    HalfBitDelay

	return
;-------------------------------
CheckSerCmd             ;[cr] terminator has been received, so
                        ;  decode the commands in here.  Will be in Stack1
    movlw   "R"         ;Rxxxx  R register data data
    subwf   Stack6, W   ;Stack6/5/4/3/2/1
    btfsc   STATUS, Z   ;     R x x x x [cr]
    goto    RCmd  

    movlw   "N"         ;Nxxxx  New N register data
    subwf   Stack6, W   ;
    btfsc   STATUS, Z   ;Stack6/5/4/3/2/1
    goto    NCmd        ;     N x x x x [cr]

    movlw   "C"         ;Cxx     New S register data
    subwf   Stack4, W   ;
    btfsc   STATUS, Z   ;Stack4/3/2/1
    goto    CCmd        ;     C x x [cr]

    movlw   "U"         ;Update Chip and continue
    subwf   Stack2, W   ;Stack2/1
    btfsc   STATUS, Z   ;
    goto    UCmd        ;     U [cr]

    movlw   "W"         ;Write to EE and continue
    subwf   Stack2, W   ;Stack2/1
    btfsc   STATUS, Z   ;
    goto    WCmd        ;     W [cr]

    goto    DoneCmd
;.............
RCmd
    movf    Stack5 , W      ;  convert to binary in R1/0
    call    DecodeHex       ;MS Nibble
    movwf   R1              ;Stack 6/5/4/3/2/1
    swapf   R1              ;      R x x x x [cr]

    movf    Stack4 , W
    call    DecodeHex
    addwf   R1
    bcf     R1, 7           ;Mask off any MSB entered in error

    movf    Stack3 , W
    call    DecodeHex
    movwf   R0
    swapf   R0

    movf    Stack2 , W      ;LSNibble
    call    DecodeHex
    addwf   R0

    movlw   0x0D
    call    SendByte
    movlw   0x0A
    call    SendByte

    movlw   "R"
    call    SendByte
    movlw   "-"
    call    SendByte

    swapf   R1 , W          ;Return data on serial link
    call    SendHex
    movf    R1 , W
    call    SendHex
    swapf   R0 , W
    call    SendHex
    movf    R0 , W
    call    SendHex

    movlw   0x0D
    call    SendByte
    movlw   0x0A
    call    SendByte         ;R1/0 updated, but not written to EE

    goto    DoneCmd
;.............
NCmd
 
    movf    Stack5 , W	    ;  convert to binary in N2/1/0
    call    DecodeHex       ;MS Nibble
    movwf   N1              ;Stack 6/5/4/3/2/1
    swapf   N1              ;      N x x x x [cr]

    movf    Stack4 , W
    call    DecodeHex
    addwf   N1

    movf    Stack3 , W
    call    DecodeHex
    movwf   N0
    swapf   N0

    movf    Stack2 , W      ;LSNibble
    call    DecodeHex
    addwf   N0

    movlw   0x0D
    call    SendByte
    movlw   0x0A
    call    SendByte

    movlw   "N"
    call    SendByte
    movlw   "-"
    call    SendByte

    swapf   N1 , W         
    call    SendHex
    movf    N1 , W
    call    SendHex
    swapf   N0 , W
    call    SendHex
    movf    N0 , W
    call    SendHex

    movlw   0x0D
    call    SendByte
    movlw   0x0A
    call    SendByte  

    goto    DoneCmd
;.............
CCmd
    movf    Stack3 , W
    call    DecodeHex
    movwf   Creg	;Stack 4/3/2/1
    swapf   Creg	;      C x x [cr]

    movf    Stack2 , W     
    call    DecodeHex
    addwf   Creg

    movlw   0x0D
    call    SendByte
    movlw   0x0A
    call    SendByte

    movlw   "C"
    call    SendByte
    movlw   "-"
    call    SendByte

    swapf   Creg , W
    call    SendHex
    movf    Creg , W
    call    SendHex

    movlw   0x0D
    call    SendByte
    movlw   0x0A
    call    SendByte  

    goto    DoneCmd
;.............
UCmd
    call    SendBack
    call    SendData

    movlw   "U"
    call    SendByte
    movlw   "p"
    call    SendByte
    movlw   "d"
    call    SendByte
    movlw   "a"
    call    SendByte
    movlw   "t"
    call    SendByte
    movlw   "e"
    call    SendByte
    movlw   "d"
    call    SendByte
    call    CrLf
    goto    DoneCmd
;..............
WCmd
    call    SendBack        ;Send back contents of actual R and N registers
    call    SendData

    call    WriteEE
    movlw   "W"
    call    SendByte
    movlw   "r"
    call    SendByte
    movlw   "i"
    call    SendByte
    movlw   "t"
    call    SendByte
    movlw   "t"
    call    SendByte
    movlw   "e"
    call    SendByte
    movlw   "n"
    call    SendByte
    call    CrLf
    goto    DoneCmd
;..............
;.............
DoneCmd
    call    ClearStack      ;Prevents spurious commands
    return
;------------------------------- 
Assemble232             ;Get here after Port B interrupt
    call    HalfBitDelay

 if RS232POL == 1
    btfss   SerIn     
 else
    btfsc   Serin
 endif

    goto    GetOut232   ;Invalid start bit

    movlw   d'8'        ;27 clocks after edge to here
    movwf   Counter
    clrf    RxChar
Rx232Loop
    call    HalfBitDelay
    call    HalfBitDelay
    bcf     STATUS , C

 if RS232POL == 1
    btfss   SerIn
 else
    btfsc   SerIn
 endif

    bsf     STATUS , C
    rrf     RxChar            
    decfsz  Counter
    goto    Rx232Loop   ;this loop 2 * HalfBitDelay + 9
    call    HalfBitDelay ;Part into stop bit - should be no more transitions


    movlw   "a"             
    subwf   RxChar, W       ;W = Char - "a"
    btfss   STATUS, C       ;if-ve (C=0) char < "a" so ignore
    goto    UpCaseDone

    movf    RxChar, W
    sublw   "z"             ;W = "z" - Char
    btfss   STATUS, C       ;if -ve (C=0)  char > "z" so ignore
    goto    UpCaseDone

    bcf     RxChar, 5       ;Reset bit 5 to force upper case
UpCaseDone
    movf    RxChar, W
    call    PushStack       ;RxChar >> Stack1, previous 5 chars move down one
    movf    RxChar , w      ;Returns with data in W as well as on stack
    
GetOut232
    return                  ;received byte in RxChar and W
;-------------------------
HalfBitDelay            ;1200 baud needs 833us, so delay 414 clocks here
    movlw   BAUDCONST   ; for a total of 414 * 2 + 8 = 833 clocks
	movwf   DelCount
BaudDelLoop             ;Total delay = 3.N + 5 = 414
	decfsz  DelCount
	goto    BaudDelLoop
	return
;-------------------------------
LongDelay
	movlw   d'150'      ;Approx 300ms delay
	movwf   Temp2
LongDelLoop1
	call	Delay
	call	Delay
	decfsz	Temp2
	goto	LongDelLoop1
	return
;------------------------------------
Delay                       ;1ms delay
    movlw   d'199'  
    movwf   DelCount
DelLoop
    nop
    nop
    decfsz  DelCount
    goto    DelLoop
    return
;------------------------------------
WriteToEE
    clrf    AutoEEADR
    movf    R1, W
    call    StoreEE_12F

    incf    AutoEEADR
    movf    R0, W
    call    StoreEE_12F

    incf    AutoEEADR
    movf    N1, W
    call    StoreEE_12F

    incf    AutoEEADR
    movf    N0, W
    call    StoreEE_12F

   incf     AutoEEADR
    movf    Creg, W
    call    StoreEE_12F

    return
;------------------------------------
GetEE_12F                   ;Enter with address in W
    bsf     STATUS, RP0
    movwf   EEADR
    bsf     EECON1, RD
    movf    EEDATA, W
    bcf     STATUS, RP0
    return                  ;Exit with data in W
;-----------------------------------------
WriteEE
    clrf    AutoEEADR
    movf    R1, W
    call    StoreEE_12F

    incf    AutoEEADR
    movf    R0, W
    call    StoreEE_12F

    incf    AutoEEADR
    movf    N1, W
    call    StoreEE_12F

    incf    AutoEEADR
    movf    N0, W
    call    StoreEE_12F

    incf    AutoEEADR
    movf    Creg, W
    call    StoreEE_12F

    return
;-------------------------------------------
StoreEE_12F        ;Enter with W = EE data,  AutoEEADR = EE address
    bsf     STATUS, RP0      ;All EE addresses in bank 1
    movwf   EEDATA
    bcf     STATUS, RP0
    movf    AutoEEADR, W
    bsf     STATUS, RP0
    movwf   EEADR    

    bsf     EECON1 , WREN    ;Enable EEprom writing
    movlw   0x55
    movwf   EECON2
    movlw   0xAA
    movwf   EECON2
    bsf     EECON1 , WR

    bcf     STATUS , RP0
EEwaitW
    btfss   PIR1 , EEIF
    goto    EEwaitW

    bcf     PIR1 , EEIF      ;has to be cleared manually
    return
;-------------------------
PushStack
    movf    Stack7 , W
    movwf   Stack8              ;Stack8 contains oldest data
    movf    Stack6 , W
    movwf   Stack7
    movf    Stack5 , W
    movwf   Stack6          
    movf    Stack4 , W
    movwf   Stack5
    movf    Stack3 , W
    movwf   Stack4
    movf    Stack2 , W
    movwf   Stack3
    movf    Stack1 , W
    movwf   Stack2
    movf    RxChar , W
    movwf   Stack1
    return
;-------------------------
ClearStack
    movlw   0
    movwf   Stack8
    movwf   Stack7
    movwf   Stack6
    movwf   Stack5
    movwf   Stack4
    movwf   Stack3
    movwf   Stack2
    movwf   Stack1
    return
;------------------------------------
Init170
    	bcf     Clk170
    	bsf     En170
    	bcf     Data170

    	movlw   d'200'          ;First start up delay
    	movwf   Counter
StartLoop1
    	call    Delay
    	decfsz  Counter
    	goto    StartLoop1

    	bsf     GPIO, 1

	bsf     Clk170	
	bcf     Clk170		;4 clocks with EN High and Din low
	bsf     Clk170
	bcf     Clk170
	bsf     Clk170
	bcf     Clk170
	bsf     Clk170
	bcf     Clk170	   	;3 clocks with En = low, D = low

	bcf     En170
	bsf     Clk170
	bcf     Clk170
	bsf     Clk170
	bcf     Clk170
	bsf     Clk170
	bcf     Clk170

    	bsf     Data170         ;1 clock with En = low, D = high
	bsf     Clk170
	bcf     Clk170
	bcf     Data170		;1 clock with D low again
	bsf     Clk170
	bcf     Clk170
	bsf     En170

    	movlw   d'100'
    	movwf   Counter
StartLoop2
    	call    Delay
    	decfsz  Counter
    	goto    StartLoop2

	return
;---------------------------------------

    cblock 0x20         ;64 Bytes available
        Wreg
        StsReg
        Temp
        Temp2
        BitCount
        Counter
        DelCount
        AutoEEADR
        D0
        D1
        D2
	Creg
        R0
        R1
        N0
        N1
        A0
        A1

        RxChar
        Stack1
        Stack2
        Stack3
        Stack4
        Stack5
        Stack6
        Stack7
        Stack8
    endc
;----------------------
	end

